import os
import sys
import pdb
import cv2
import numpy
import theano

from pylearn2.datasets.dataset import Dataset
from pylearn2.datasets.dense_design_matrix import DenseDesignMatrix
from pylearn2.datasets.dense_design_matrix import DefaultViewConverter
from pylearn2.datasets import control
from pylearn2.space import CompositeSpace, Conv2DSpace, VectorSpace, IndexSpace
from pylearn2.utils import safe_zip
from pylearn2.utils.rng import make_np_rng
from theano import config

def normalize(X, low, high, dtype=None):
	"""Normalizes a given array in X to a value between low and high."""
	X = numpy.asarray(X)
	minX, maxX = numpy.min(X), numpy.max(X)
	# normalize to [0...1].
	X = X - float(minX)
	X = X / float((maxX - minX))
	# scale to [low...high].
	X = X * (high-low)
	X = X + low
	if dtype is None:
		return numpy.asarray(X)
	return numpy.asarray(X, dtype=dtype)

def label_to_dir(path, label, splitsize=None, splitstart=None, nclasses=None, sz=None, extension=None):
	c = 0    
	ctr1 = 0
	for dirname, dirnames, filenames in os.walk(path):		
		for subdirname in dirnames:
			if nclasses is not None and ctr1 >= nclasses:
				break
			ctr1 += 1
			ctr2 = 0
			subject_path = os.path.join(dirname, subdirname)
			for filename in os.listdir(subject_path):
				if filename.startswith(".") or (extension is None and not (filename.endswith(".png") or filename.endswith(".pgm"))) or (extension is not None and not filename.endswith(extension)):
					continue                
				if splitstart is not None and ctr2 < splitstart:
					ctr2 += 1
					continue
				if splitsize is not None and splitstart is not None and ctr2 >= splitsize+splitstart:
					break
				ctr2 += 1
			if label == c:
				return os.path.join(dirname, subdirname)
			c = c+1
	return None


def read_data_dir(path, splitsize=None, splitstart=None, nclasses=None, splitpct=None, sz=None, grayscale=True, extension=None):
	"""Reads the images in a given folder, resizes images on the fly if size is given.

	Args:
		path: Path to a folder with subfolders representing the subjects (persons).
		splitsize: Number of images from each directory to read
		splitstart: Offset within each directory to start reading images
		nclasses: Number of image subdirectories (classes) to read
		sz: A tuple with the size Resizes

	Returns:
		A list [X,y]

			X: The images, which is a Python list of numpy arrays.
			y: The corresponding labels (the unique number of the subject, person) in a Python list.
	"""
	c = 0
	X,y = [], []
	donedirs = []
	ctr1 = 0
	for dirname, dirnames, filenames in os.walk(path):		
		if len(dirnames) == 0 and dirname not in donedirs:
			dirnames = [""]		
		for subdirname in dirnames:
			if nclasses is not None and ctr1 >= nclasses:
				break

			ctr1 += 1
			ctr2 = 0
			subject_path = os.path.join(dirname, subdirname)
			relevant_files = [filename for filename in os.listdir(subject_path) if not (filename.startswith(".") or (extension is None and not (filename.endswith(".png") or filename.endswith(".pgm"))) or (extension is not None and not filename.endswith(extension)))]

			donedirs.append(subject_path)

			if splitpct is not None:
				first_splitsize = int(float(len(relevant_files))*splitpct) # round down so we always have non-zero splits if splitpct < 1.0
				if splitstart > 0:
					splitsize = len(relevant_files) - first_splitsize
					splitstart = first_splitsize
				else:
					splitsize = first_splitsize

			for filename in relevant_files:
				if splitstart is not None and ctr2 < splitstart:
					ctr2 += 1
					continue
				if splitsize is not None and splitstart is not None and ctr2 >= splitsize+splitstart:
					break
				ctr2 += 1               
				try:
					if grayscale:
						im = cv2.imread(os.path.join(subject_path, filename), cv2.IMREAD_GRAYSCALE)
					else:
						im = cv2.imread(os.path.join(subject_path, filename))                        
					# resize to given size (if given)
					if (sz is not None):
						im = cv2.resize(im, sz)                        

					# print "adding %s with label %d" % (filename, c)
					X.append(normalize(numpy.asarray(im, dtype=numpy.uint8).flatten(), 0, 1, theano.config.floatX))                 
					y.append(c)
				except IOError, (errno, strerror):
					print "I/O error({0}): {1}".format(errno, strerror)
				except:
					print "Unexpected error:", sys.exc_info()[0]
					raise
			c = c+1	
	return [numpy.asarray(X, dtype=theano.config.floatX), numpy.asarray(y, dtype=numpy.int32)]

def data_dir_shape(path, extension):
	for dirname, dirnames, filenames in os.walk(path):
		if len(dirnames) == 0:
			dirnames = [""]        
		for subdirname in dirnames:
			subject_path = os.path.join(dirname, subdirname)
			for filename in os.listdir(subject_path):
				if filename.startswith(".") or (extension is None and not (filename.endswith(".png") or filename.endswith(".pgm"))) or (extension is not None and not filename.endswith(extension)):
					continue
				try:
					im = cv2.imread(os.path.join(subject_path, filename), cv2.IMREAD_GRAYSCALE)                 
					return im.shape
				except IOError, (errno, strerror):
					print "I/O error({0}): {1}".format(errno, strerror)
				except:
					print "Unexpected error:", sys.exc_info()[0]
					raise
	raise Exception('Could not get shape in %s' % path)

class DataDirViewConverter(object):
	"""
	.. todo::

		WRITEME

	Parameters
	----------
	shape : WRITEME
	axes : WRITEME
	"""
	def __init__(self, shape, axes=('b', 0, 1, 'c')):
		self.shape = shape
		self.pixels_per_channel = 1
		for dim in self.shape[:-1]:
			self.pixels_per_channel *= dim
		self.axes = axes
		self._update_topo_space()

	def view_shape(self):
		"""
		.. todo::

			WRITEME
		"""
		return self.shape

	def weights_view_shape(self):
		"""
		.. todo::

			WRITEME
		"""
		return self.shape

	def design_mat_to_topo_view(self, X):
		"""
		.. todo::

			WRITEME
		"""
		assert len(X.shape) == 2
		batch_size = X.shape[0]
		nchannels = self.shape[-1]
		dimshuffle_args = [('b', 0, 1, 'c').index(axis) for axis in self.axes]
		new_shape = [batch_size, self.shape[0], self.shape[1], nchannels]

		return X.reshape(new_shape).transpose(*dimshuffle_args)

	def design_mat_to_weights_view(self, X):
		"""
		.. todo::

			WRITEME
		"""
		rval = self.design_mat_to_topo_view(X)

		# weights view is always for display
		rval = np.transpose(rval, tuple(self.axes.index(axis)
										for axis in ('b', 0, 1, 'c')))

		return rval

	def topo_view_to_design_mat(self, V):
		"""
		.. todo::

			WRITEME
		"""

		V = V.transpose(self.axes.index('b'),
						self.axes.index(0),
						self.axes.index(1),
						self.axes.index('c'))

		num_channels = self.shape[-1]
		if np.any(np.asarray(self.shape) != np.asarray(V.shape[1:])):
			raise ValueError('View converter for views of shape batch size '
							 'followed by ' + str(self.shape) +
							 ' given tensor of shape ' + str(V.shape))
		batch_size = V.shape[0]

		rval = np.zeros((batch_size, self.pixels_per_channel * num_channels),
						dtype=V.dtype)

		for i in xrange(num_channels):
			ppc = self.pixels_per_channel
			rval[:, i * ppc:(i + 1) * ppc] = V[..., i].reshape(batch_size, ppc)
		assert rval.dtype == V.dtype

		return rval

	def get_formatted_batch(self, batch, dspace):
		"""
		.. todo::

			WRITEME properly

		Reformat batch from the internal storage format into dspace.
		"""
		if isinstance(dspace, VectorSpace):
			# If a VectorSpace is requested, batch should already be in that
			# space. We call np_format_as anyway, in case the batch needs to be
			# cast to dspace.dtype. This also validates the batch shape, to
			# check that it's a valid batch in dspace.
			return dspace.np_format_as(batch, dspace)
		elif isinstance(dspace, Conv2DSpace):
			# design_mat_to_topo_view will return a batch formatted
			# in a Conv2DSpace, but not necessarily the right one.
			topo_batch = self.design_mat_to_topo_view(batch)
			if self.topo_space.axes != self.axes:
				warnings.warn("It looks like %s.axes has been changed "
							  "directly, please use the set_axes() method "
							  "instead." % self.__class__.__name__)
				self._update_topo_space()

			return self.topo_space.np_format_as(topo_batch, dspace)
		else:
			raise ValueError("%s does not know how to format a batch into "
							 "%s of type %s."
							 % (self.__class__.__name__, dspace, type(dspace)))

	def __setstate__(self, d):
		"""
		.. todo::

			WRITEME
		"""
		# Patch old pickle files that don't have the axes attribute.
		if 'axes' not in d:
			d['axes'] = ['b', 0, 1, 'c']
		self.__dict__.update(d)

		# Same for topo_space
		if 'topo_space' not in self.__dict__:
			self._update_topo_space()

	def _update_topo_space(self):
		"""Update self.topo_space from self.shape and self.axes"""
		rows, cols, channels = self.shape
		self.topo_space = Conv2DSpace(shape=(rows, cols),
									  num_channels=channels,
									  axes=self.axes)

	def set_axes(self, axes):
		"""
		.. todo::

			WRITEME
		"""
		self.axes = axes
		self._update_topo_space()


class DataDirDataset(DenseDesignMatrix):

	def __init__(self, data_dir, splitsize=None, splitstart=None, nclasses=None, splitpct=None, channels=1, grayscale=True, sizex=None, sizey=None, extension=None):

		assert (channels == 1 or not grayscale) and "grayscale images have exactly one channel"
		assert (sizex is not None and sizey is not None) or (sizex is None and sizey is None) and "must provide BOTH dimension sizes (or none)"

		size = (sizex, sizey) if sizex is not None else None        

		rows, cols = data_dir_shape(data_dir, extension) if sizex is None else (sizey, sizex)
		train_set_x, train_set_y = read_data_dir(data_dir, splitsize=splitsize, splitstart=splitstart, nclasses=nclasses, splitpct=splitpct, sz=size, grayscale=grayscale, extension=extension)		
		super(DataDirDataset, self).__init__(X=train_set_x, y=train_set_y, y_labels=nclasses, view_converter=DataDirViewConverter((rows, cols, channels)))        
		self.convert_to_one_hot()

# dd = DataDirDataset("../images/political", splitsize=5, splitstart=0, nclasses=2, channels=3, grayscale=False, sizex=92, sizey=112, extension=".png")
# dd.get_batch_topo(5)

# DataDirDataset("../images/aberdeen-mono-labeled/train", splitpct=0.5, splitstart=1, channels=1, grayscale=True, extension=".jpg")
