import os
import sys
import shutil
import csv
import random
import re
import pdb
import pickle
import tempfile
import math
import argparse

import cv2

html_template = """
<html>
<head>

<meta http-equiv="content-type" content="text/html; charset=UTF-8">
<title></title>
</head>
<body>
<div align="center"><img alt=""
src=ALTERED
height="224" width="184"><br>
Altered Image<br>
<br>
<table border="1" cellpadding="2" cellspacing="2" width="100%">
<tbody>
<tr>
<td align="center" valign="top"><img alt=""
src=FACE1
height="224" width="184"><br>
</td>
<td align="center" valign="top"><img alt=""
src=FACE2
height="224" width="184"><br>
</td>
<td align="center" valign="top"><img alt=""
src=FACE3
height="224" width="184"><br>
</td>
<td align="center" valign="top"><img alt=""
src=FACE4
height="224" width="184"><br>
</td>
<td align="center" valign="top"><img alt=""
src=FACE5
height="224" width="184"><br>
</td>
</tr>
<tr>
<td align="center" valign="top">#1<br>
</td>
<td align="center" valign="top">#2<br>
</td>
<td align="center" valign="top">#3<br>
</td>
<td align="center" valign="top">#4<br>
</td>
<td align="center" valign="top">#5<br>
</td>
</tr>
</tbody>
</table>
<br>
<ol>
<li>Face #1</li>
<li>Face #2</li>
<li>Face #3</li>
<li>Face #4</li>
<li>Face #5</li>
<li>Not Present<br>
</li>
</ol>
</div>
</body>
</html>
"""

def random_image(image_dir):

    images = os.walk(image_dir).next()[2]

    return os.path.join(image_dir, random.choice(images))

def do_blur_row(exp_dir, face_dir, use_bb, include_truth, is_control, num_tests):

    bb_str = "true" if use_bb else "false"
    bb_matcher = re.compile(".+_bb=(true|false)_s\d+")

    image_dirs = [t[0] for t in os.walk(face_dir) if t[0] != face_dir]
    exp_dirs = [t[0] for t in os.walk(exp_dir) if t[0] != exp_dir and bb_matcher.match(t[0]) is not None]

    exp = random.choice(exp_dirs)   

    label_matcher = re.compile(".+(s\d+)")
    true_label = label_matcher.match(exp).group(1)

    labels = [label_matcher.match(imdir).group(1) for imdir in image_dirs]

    tests = []
    perm = range(num_tests)
    random.shuffle(perm)    

    info = {}
    info["true_label"] = true_label
    if not is_control:
        info["inverted_face"] = os.path.join(exp, "final_image.png")
    else:
        info["inverted_face"] = os.path.join(exp, "blurred_image.png")
    
    # add a random copy of the true face to the tests
    if include_truth:
        tests.append(random_image(os.path.join(face_dir, true_label)))
        info["true_idx"] = perm[len(tests)-1]
    else:
        info["true_idx"] = -1

    info["control"] = str(is_control).lower()

    # pick a random duplicated face, add it to the tests
    # dup_label = random.choice(labels)
    # tests.append(random_image(os.path.join(face_dir, dup_label)))
    # info["dup1_idx"] = perm[len(tests)-1]
    # tests.append(random_image(os.path.join(face_dir, dup_label)))
    # info["dup2_idx"] = perm[len(tests)-1]

    # pick the remaining tests randomly
    used_labels = [true_label]
    while len(tests) < num_tests:
        label = true_label
        while label in used_labels:
            label = random.choice(labels)
        used_labels.append(label)
        tests.append(random_image(os.path.join(face_dir, label)))
    
    # permute the tests
    shuffled_tests = [None for i in range(len(tests))]
    for i in range(len(tests)):
        shuffled_tests[perm[i]] = tests[i]

    info["tests"] = shuffled_tests  

    return info

def do_row(exp_dir, face_dir, use_bb, include_truth, is_control, num_tests):

    bb_str = "true" if use_bb else "false"
    bb_matcher = re.compile(".+_bb=(true|false)_s\d+")

    image_dirs = [t[0] for t in os.walk(face_dir) if t[0] != face_dir]
    exp_dirs = [t[0] for t in os.walk(exp_dir) if t[0] != exp_dir and bb_matcher.match(t[0]) is not None]

    exp = random.choice(exp_dirs)   

    label_matcher = re.compile(".+(s\d+)")
    true_label = label_matcher.match(exp).group(1)

    labels = [label_matcher.match(imdir).group(1) for imdir in image_dirs]

    tests = []
    perm = range(num_tests)
    random.shuffle(perm)    

    info = {}
    info["true_label"] = true_label
    if not is_control:
        info["inverted_face"] = os.path.join(exp, "final_image.png")
    else:
        info["inverted_face"] = os.path.join(exp, "target_img.png")
    
    # add a random copy of the true face to the tests
    if include_truth:
        tests.append(random_image(os.path.join(face_dir, true_label)))
        info["true_idx"] = perm[len(tests)-1]
    else:
        info["true_idx"] = -1

    info["control"] = str(is_control).lower()

    # pick a random duplicated face, add it to the tests
    # dup_label = random.choice(labels)
    # tests.append(random_image(os.path.join(face_dir, dup_label)))
    # info["dup1_idx"] = perm[len(tests)-1]
    # tests.append(random_image(os.path.join(face_dir, dup_label)))
    # info["dup2_idx"] = perm[len(tests)-1]

    # pick the remaining tests randomly
    used_labels = [true_label]
    while len(tests) < num_tests:
        label = true_label
        while label in used_labels:
            label = random.choice(labels)
        used_labels.append(label)
        tests.append(random_image(os.path.join(face_dir, label)))
    
    # permute the tests
    shuffled_tests = [None for i in range(len(tests))]
    for i in range(len(tests)):
        shuffled_tests[perm[i]] = tests[i]

    info["tests"] = shuffled_tests  

    return info

def write_png(input_file, output_file):

    if not output_file.endswith(".png"):
        output_file = "%s.png" % output_file

    img = cv2.resize(cv2.imread(input_file), (184, 224))
    cv2.imwrite(output_file, img)   

def csvheader(info):

    feats = ["true_label", "true_idx", "control", "inverted_face"]
    for i in range(len(info["tests"])):
        feats.append("test%d" % (i+1))

    return (feats, ",".join(feats))

def unique_filename(prefix="img", dir="."):

    fd, name = tempfile.mkstemp(suffix=".png", prefix=prefix, dir=dir)
    os.close(fd)
    
    return name

def do_cache(cache, image_file, image_output_dir):

    if cache.has_key(image_file):
    
        cached = cache[image_file]
    
    else:
    
        newname = unique_filename(dir=image_output_dir)
        write_png(image_file, newname)
        cache[image_file] = newname

        cached = newname

    basename = os.path.basename(cached)

    return "https://raw.githubusercontent.com/mfredrik/facematch/master/%s" % basename

def row2csv(info, face_dir, image_output_dir):

    cachefile = os.path.join(image_output_dir, "namecache.pkl")
    if not os.path.exists(cachefile):
        cache = {}
    else:
        with open(cachefile, "rb") as cachefp:
            cache = pickle.load(cachefp)

    try:
        info["inverted_face"] = do_cache(cache, info["inverted_face"], image_output_dir)

        (feats, _) = csvheader(info)
        vals = [str(info[feat]) if not isinstance(info[feat], int) else str(info[feat]+1) for feat in feats if info.has_key(feat)]
        for i in range(len(info["tests"])):
            info["tests"][i] = do_cache(cache, info["tests"][i], image_output_dir)
            vals.append(str(info["tests"][i]))
    finally:
        with open(cachefile, "wb") as cachefp:
            pickle.dump(cache, cachefp)

    return (feats, ",".join(vals))

def print_csv(feats, rows, fp=sys.stdout):

    fp.write(",".join(feats) + "\n")    
    for row in rows:
        fp.write(row + "\n")

def print_hit(info, fp=sys.stderr):

    html = html_template[:]

    html = html.replace("ALTERED", "\"%s\"" % info["inverted_face"])
    for i in range(len(info["tests"])):
        html = html.replace("FACE%d" % (i+1), "\"%s\"" % info["tests"][i])

    fp.write(html + "\n")

def process_args():

  parser = argparse.ArgumentParser(description='Image inversion turk experiment generator.')
  parser.add_argument('exp_dir', metavar='exp_dir', type=str, help='top-level experiment directory')
  parser.add_argument('face_dir', metavar='face_dir', type=str, help='top-level face image directory')
  parser.add_argument('face_output_dir', metavar='face_output_dir', type=str, help='directory to cache face images')
  parser.add_argument('n', metavar='n', type=int, help='number of HITs to generate')
  parser.add_argument('csv', metavar='csv', type=str, help='csv output filename')
  parser.add_argument('truth_pct', metavar='truth_pct', type=float, help='fraction of HITs containing original image')
  parser.add_argument('control_pct', metavar='control_pct', type=float, help='fraction of HITs in the control group')
  parser.add_argument('--deblur', action='store_true', help='flag for facial deblurring experiments')

  return parser.parse_args()

def main():

    args = process_args()

    random.seed(4)

    csvlines = []

    exp_dir = args.exp_dir
    face_dir = args.face_dir
    face_output_dir = args.face_output_dir
    n = args.n
    out_csv = args.csv
    truth_pct = args.truth_pct
    control_pct = args.control_pct
    do_blur = args.deblur

    rows = []

    if truth_pct > 0.0:
        truth_test_n = int(max(round(truth_pct*float(n)*(1.0-control_pct)), 1))
    else:
        truth_test_n = 0
    if control_pct > 0.0:
        truth_control_n = int(truth_test_n + max(round(truth_pct*float(n)*control_pct), 1))
    else:
        truth_control_n = 0
    if truth_pct < 1.0:
        notruth_test_n = int(min(truth_control_n + round((1.0 - truth_pct)*float(n)*(1.0-control_pct)), n-1))
    else:
        notruth_test_n = n+1    

    include_truth = True
    is_control = False        

    for i in range(n):

        if i == truth_test_n:
            is_control = True
        if i == truth_control_n:
            include_truth = False
            is_control = False
        if i == notruth_test_n:
            is_control = True

        if not do_blur:
            row = do_row(exp_dir, face_dir, use_bb=False, include_truth=include_truth, is_control=is_control, num_tests=5)
        else:
            row = do_blur_row(exp_dir, face_dir, use_bb=False, include_truth=include_truth, is_control=is_control, num_tests=5)
        feats, vals = row2csv(row, face_dir, face_output_dir)
        with open("/tmp/hit%d.html" % (i+1), "w") as fp:
            print_hit(row, fp)
        rows.append(vals)
    
    with open(out_csv, "w") as fp:
        print_csv(feats, rows, fp)

if __name__ == '__main__':
    main()
