import sys
import os
import argparse
import gzip
import math
import random
import subprocess as sp
import time

import pylearn2
import theano
import scipy
import numpy as np
import cv2
import cPickle

from pylearn2.config import yaml_parse
from pylearn2.utils import serial
from pylearn2.utils import safe_zip
from pylearn2.utils.data_specs import DataSpecsMapping

from PIL import Image

def normalize(X, low, high, dtype=theano.config.floatX):
	"""Normalizes a given array in X to a value between low and high."""
	X = np.asarray(X)
	# minX, maxX = max(np.min(X), 1), max(np.max(X), 1)
	minX, maxX = np.min(X), np.max(X)
	if minX == maxX:
		minX = low
		maxX = high
	# print minX, ",", maxX
	# normalize to [0...1].
	X = X - float(minX)
	X = X / float((maxX - minX))
	# scale to [low...high].
	X = X * (high-low)
	X = X + low
	if dtype is None:
		return np.asarray(X)
	return np.asarray(X, dtype=dtype)

class EncoderCascade(object):

	def __init__(self, encoders):

		self.encoders = encoders
		self.nvis = encoders[0].input_space.dim
		self.nhid = encoders[-1].output_space.dim

	def encode_expr(self, input_expr):

		expr = input_expr
		for enc in self.encoders:
			expr = enc.encode(expr)

		return expr

	def decode_expr(self, input_expr):

		expr = input_expr
		for enc in self.encoders[::-1]:
			expr = enc.decode(expr)

		return expr

	def encode(self, input_expr):

		if not isinstance(input_expr, theano.tensor.Variable):
			input_expr = theano.shared(value=input_expr, name='input')

		encode_expr = self.encode_expr(input_expr)
		f = theano.function([], encode_expr)

		return f()

	def decode(self, input_expr):

		if not isinstance(input_expr, theano.tensor.Variable):
			input_expr = theano.shared(value=input_expr, name='input')

		decode_expr = self.decode_expr(input_expr)
		f = theano.function([], decode_expr)

		return f()

def pixelate_img(orig_img, npix=10):

	height, width = orig_img.shape
	img = np.copy(orig_img)

	for x in range(0, height, npix):
		for y in range(0, width, npix):
			ux = min(x+npix, height)
			uy = min(y+npix, width)
			img[x:ux, y:uy] = np.mean(orig_img[x:ux, y:uy])			

	return img

def sharpen(image, iterations=1):
	sharpened = np.copy(image)
	for i in range(iterations):
		blurred = cv2.GaussianBlur(sharpened, (3, 3), 0)
		sharpened = cv2.addWeighted(sharpened, 1.5, blurred, -0.5, 0)

	return sharpened

if __name__ == '__main__':

	# image_file = sys.argv[1]
	encoder_files = sys.argv[2:]

	# print

	t_start = time.time()
	# orig_image = cv2.imread(image_file, cv2.IMREAD_GRAYSCALE)
	# image = pixelate_img(orig_image, 25)
	# image = orig_image
	encoders = [serial.load(encoder_file) for encoder_file in encoder_files]
	cascade = EncoderCascade(encoders)
	with open(sys.argv[1], "wb") as fp:
		cPickle.dump(cascade, fp)
	# t_stop = time.time()	

	# print "... finished loading encoders (%f seconds)" % (t_stop - t_start)

	# image_v = theano.shared(value=normalize(image.reshape(cascade.nvis), 0, 1), name='image')
	# encoded = cascade.encode(image_v)
	# encoded_v = theano.shared(value=encoded, name='encoded')
	# decoded = normalize(cascade.decode(encoded_v).reshape(image.shape), 0, 255, dtype=np.uint8)
	# sharpened = sharpen(decoded, 2)

	# print "diff:", np.sum(np.abs(decoded - sharpened))
	# print "... done. press any key while focused on results window to terminate."

	# display_image = np.concatenate((orig_image, image, decoded, sharpened), axis=1)
	# display_image = cv2.resize(display_image, None, fx=2, fy=2, interpolation=cv2.INTER_NEAREST)
	# cv2.namedWindow('results', cv2.CV_WINDOW_AUTOSIZE)
	# cv2.imshow('results', display_image)
	# cv2.waitKey(0)
	
