import sys
import pprint
import math
import csv
import json
import pdb
import numpy
import random

import treeinversion

import Orange as orange

from nltk.metrics import ConfusionMatrix
from sklearn import metrics
from sklearn.cross_validation import StratifiedShuffleSplit

pp = pprint.PrettyPrinter(indent=2)

class OrangeModelInfo(treeinversion.ModelInfo):

	def __init__(self, model_object):

		self.training_size = model_object["max_rows"] # yes, this is overwritten. not sure which to use yet
		self.objective_field = model_object["objective_field"]

		obj = model_object["model"]
		self.field_names = [obj["fields"][fkey]["name"] for fkey in obj["fields"].keys()]
		self.field_values = {obj["fields"][fkey]["name"]: [cat[0] for cat in obj["fields"][fkey]["summary"]["categories"]] for fkey in obj["fields"].keys()}
		self.training_size = sum([cat[1] for cat in obj["distribution"]["training"]["categories"]])
		self.marginals = {obj["fields"][fkey]["name"]: {cat[0]: float(cat[1])/float(self.training_size) for cat in obj["fields"][fkey]["summary"]["categories"]} for fkey in obj["fields"].keys()}
		self.name_to_id = {obj["fields"][fkey]["name"]: fkey for fkey in obj["fields"].keys()}
		self.id_to_name = {fkey: obj["fields"][fkey]["name"] for fkey in obj["fields"].keys()}
		self.objective_name = self.id_to_name[self.objective_field]
		self.objective_index = sum([i for i in range(len(self.field_names)) if self.field_names[i] == self.objective_name])
		self.root = obj["root"]
		self.counter = [0]

def onode2bnode(node, pred, field_ids):

	bnode = {}

	if node is None:
		bnode["count"] = 0
		bnode["predicate"] = pred
		bnode["output"] = None
		bnode["objective_summary"] = {"categories": []}

		return bnode

	dist = node.distribution
	cats = [list(el) for el in zip(dist.keys(), dist.values())]
	bnode["objective_summary"] = {"categories": cats}

	bnode["count"] = int(sum(dist.values()))
	bnode["predicate"] = pred
	bnode["output"] = dist.modus().value

	if node is None:
		return bnode

	if node.branches is None:
		return bnode
	else:
		bnode["children"] = []
		var = node.branch_selector.class_var.name
		assert len(node.branches) == len(node.branch_sizes)
		for (val, i) in zip(node.branch_descriptions, range(len(node.branches))):
		# for (val, i) in zip(node.branch_sizes.keys(), range(len(node.branches))):
			if val.startswith("not "):
				operator = "!="
				val = val[4:]
			else:
				operator = "="

			# print node.branch_descriptions[i], node.branches[i]
			newpred = {"operator": operator, "field": field_ids[var], "value": val}
			bnode["children"].append(onode2bnode(node.branches[i], newpred, field_ids))

		return bnode

def is_number(s):
	try:
		x = float(s)
	except:
		return False
	return True

def roundto(x, base=10):
    return int(base * round(float(x)/base))

def csv2orangetable(filename, classattr, target):

	data = []
	features = []
	values = {}

	with open(filename, 'rU') as f:
		reader = csv.reader(f, delimiter = ",", quotechar = "\"")
		attrline = reader.next()
		attrline = [el.strip() for el in attrline]
		features = attrline[:]
		attr_indices = {attrline[i]: i for i in range(len(attrline))}
		attributes = {i: attrline[i] for i in range(len(attrline))}
		values = {feature: set() for feature in attrline}
		types = {feature: "discrete" for feature in attrline}

		for row in reader:
			drow = []
			classval = None
			missing_target = False
			for i in range(len(attributes)):
				missing = len(str(row[i]).strip()) == 0
				missing_target = missing_target or (missing and (attributes[i] == target or attributes[i] == classattr))
				if missing_target:
					break
				if is_number(row[i]):
					types[attributes[i]] = "numeric"
				if attributes[i] != classattr:
					if len(row[i].strip()) == 0:
						row[i] = None
					else:
						if types[attributes[i]] == "numeric":
							base = 10 if attributes[i] == "Age" else 1
							row[i] = float(row[i]) if len(row[i].strip()) == 0 else None
					drow.append(row[i])
				else:
					classval = row[i]
				if row[i] is not None:
					if types[attributes[i]] == "numeric":
						base = 10 if attributes[i] == "Age" else 1
						row[i] = roundto(row[i], base)
					values[attributes[i]].update([row[i]])

			if not missing_target:
				assert classval is not None
				drow.append(classval)
				data.append(drow)

	noclass = [feature for feature in features if feature != classattr]
	feats = []
	for name in noclass:
		if types[name] == "discrete":
			feats.append(orange.feature.Discrete(name, values=list(values[name])))
		else:
			feats.append(orange.feature.Continuous(name))
	classfeat = orange.feature.Discrete(classattr, values=list(values[classattr]))
	dom = orange.data.Domain(feats, classfeat)

	table = orange.data.Table(dom, data)
	print "read %d rows" % len(table)

	return table, values, [feat for feat in features if feat != classattr]

def orangetable2bigml(table):

	bmltable = []
	dom = table.domain

	for row in table:
		bmlrow = {}
		for domfield in dom:
			bmlrow[domfield.name] = row[domfield].value

		bmltable.append(bmlrow)

	return bmltable

def orange2bigml(tree, data, values):

	def bmlrow2orange(row, domain):
		if not isinstance(row, dict):
			return row
		orow = []
		for feat in domain:
			orow.append(row[feat.name])
		return orange.data.Instance(domain, orow)

	bml_model = lambda d: tree(bmlrow2orange(d, tree.domain)).value

	bml = {}
	model = {}

	fields = [dom.name for dom in tree.domain]
	field_ids = {tree.domain[i].name: str(i) for i in range(len(tree.domain))}
	ids_to_fields = {str(i): tree.domain[i].name for i in range(len(tree.domain))}

	model["root"] = onode2bnode(tree.tree, True, field_ids)

	bml["max_rows"] = sum(tree.tree.distribution.values())
	bml["objective_field"] = field_ids[tree.class_var.name]
	bml["model"] = model

	model["fields"] = {field_ids[dom.name]: {"name": dom.name, "summary": {}} for dom in tree.domain}
	# distrs = {dom.name: {val: 0 for val in dom.values} for dom in data.domain}
	distrs = {dom.name: {val: 0 for val in values[dom.name]} for dom in data.domain}
	p_distrs = {val: 0 for val in tree.tree.distribution.keys()}
	for row in data:
		for dom in data.domain:
			if row[dom.name].value == "?":
				continue
			distrs[dom.name][row[dom.name].value] += 1
		p_distrs[bml_model(row)] += 1
	for field_id in model["fields"]:
		cats = []
		summ = model["fields"][field_id]["summary"]
		summ["categories"] = cats
		summ["missing_count"] = 0

		field = ids_to_fields[field_id]
		for val in distrs[field].keys():
			cats.append([val, int(distrs[field][val])])

	model["distribution"] = {}
	model["distribution"]["training"] = {"categories": model["fields"][bml["objective_field"]]["summary"]["categories"]}
	model["distribution"]["predictions"] = {"categories": [[val, int(p_distrs[val])] for val in p_distrs.keys()]}

	return bml, bml_model

class SplitConstructor_Experimental(object):
	
	def __init__(self, scons, features, target, rand = None):
		self.scons = scons           # split constructor of original tree
		# self.attributes = attributes # number of features to consider
		self.rand = rand
		self.features = features
		self.target = target
		self.targetidx = sum([i if features[i] == target else 0 for i in range(len(features))])
		self.is_first = True
		if not self.rand:
			self.rand = random.Random(0)

	def __call__(self, gen, weight_id, contingencies, apriori, candidates, clsfr):

		# candidates[self.targetidx] = 0
		if self.is_first:
			candidates = [1 if i == self.targetidx else 0 for i in range(len(self.features))]
			self.is_first = False
		(sel, des, sizes, quality, spent) = self.scons(gen, weight_id, contingencies, apriori, candidates, clsfr)

		if sel is None and False:
			# pdb.set_trace()
			new_cands = [1 if i == self.targetidx else 0 for i in range(len(self.features))]
			(sel, des, sizes, quality, spent) = self.scons(gen, weight_id, contingencies, apriori, new_cands, clsfr)

		return (sel, des, sizes, quality, spent)

def train_orange_tree(config, split, shuffle=False):

	target = str(config["target"])
	classname = str(config["class"])

	if not config.has_key("seed"):
		config["seed"] = 1
	rseed = config["seed"]
	rand = random.Random(rseed)
	config["seed"] += 1

	data, values, feats = csv2orangetable(config["datafile"], classname, target)
	if config.has_key("split") and split < 1.0:
		targvals = [row[target] for row in data]
		classvals = [row[classname] for row in data]
		stratvals = ["%s,%s" % (t,c) for (t,c) in zip(targvals, classvals)]
		sss = StratifiedShuffleSplit(stratvals, 1, train_size=float(split), random_state=rseed)
		for train_inds, test_inds in sss:
			print "train size:", len(train_inds)
			training = orange.data.Table([data[i] for i in train_inds])
			test = orange.data.Table([data[i] for i in test_inds])
	else:
		if shuffle:
			rand.shuffle(data)
		training = orange.data.Table(data[:])
		test = orange.data.Table(data[:])
	bml_training = orangetable2bigml(training)
	bml_test = orangetable2bigml(test)

	learner = orange.regression.tree.TreeLearner(max_depth=1000)
	scons = learner.split
	# learner.split = SplitConstructor_Experimental(scons, feats, target)
	tree = learner(training)
	# tree.dot("/tmp/tree.dot")
	# pdb.set_trace()

	bml, bml_model = orange2bigml(tree, training, values)
	info = OrangeModelInfo(bml)

	return bml, bml_model, info, bml_training, bml_test
