import sys
import csv

def collapse_spaces(line):

	oldline = None
	while oldline != line:
		oldline = line
		line = line.replace("\t", " ")
		line = line.replace("  ", " ")

	return line

def read_ddl(filename, remove_val_commas=True, remove_val_squotes=True, remove_val_dquotes=True, replace_unknown="N/A"):

	defs = {}

	in_catlabels = False
	in_other = False
	name = None
	missing_vals = []

	with open(filename, 'rb') as f:
		for line in f:
			line = collapse_spaces(line)
			splits = [sp.strip() for sp in line.split()]
			
			if "=" in splits:
				if in_catlabels:
					in_catlabels = False
				if splits[0] == "catlabels":
					in_catlabels = True					
					defs[name] = {}
					splits = splits[2:]
				if splits[0] == "name":
					name = splits[2]
				if splits[0] == "md":
					missing_vals = [sp.strip() for sp in " ".join(splits[2:]).split(",")]

			if in_catlabels:
				val = splits[0]
				if val in missing_vals and replace_unknown is not None:
					newval = replace_unknown
				else:
					newval = " ".join(splits[1:])
					if remove_val_commas:
						newval = newval.replace(',', '')
					if remove_val_squotes:
						newval = newval.replace('\'', '')
					if remove_val_dquotes:
						newval = newval.replace('\"', '')

				defs[name][val] = newval

	# for name in defs.keys():
	# 	print name
	# 	for val in defs[name].keys():
	# 		print "\t%s --> %s" % (val, defs[name][val])
	# 	print

	return defs

def readcsv(filename, delete_with_missing=True, remove_val_commas=True, remove_val_squotes=True, remove_val_dquotes=True):

	data = []
	data_rec = []

	with open(filename, 'rb') as f:		
		# dialect = csv.Sniffer().sniff(f.read(1024), delimiters=",")
		# f.seek(0)
		dialect = csv.excel		
		reader = csv.reader(f, dialect)

		first = True
		feats = None
		for row in reader:
			if first:
				feats = [val.strip() for val in row]
			minlen = min([len(val) for val in row])
			if minlen == 0 and delete_with_missing:
				continue
			newrow = []
			newrec = {}
			for (val, i) in zip(row, range(len(row))):
				newval = val
				if remove_val_commas:
					newval = newval.replace(',', '')
				if remove_val_squotes:
					newval = newval.replace('\'', '')
				if remove_val_dquotes:
					newval = newval.replace('\"', '')
				newrow.append(newval)
				if not first:
					newrec[feats[i]] = newval			
			row = newrow

			data.append(row)
			if not first:
				data_rec.append(newrec)

			first = False

	return data, data_rec, dialect

def fit_to_def(data_rec, datadef):

	new_data_rec = []

	for row in data_rec:
		new_row = {}
		for feat in datadef.keys():
			if row[feat] in datadef[feat]:
				new_row[feat] = datadef[feat][row[feat]]
			else:
				new_row[feat] = row[feat]

		new_data_rec.append(new_row)

	return new_data_rec

def recs_to_rows(data_rec):

	feats = [feat for feat in data_rec[0].keys()]

	data = []
	data.append(feats)

	for row in data_rec:
		data.append([row[feat] for feat in feats])

	return data

def writecsv(data, filename, dialect):

	with open(filename, 'wb') as f:
		writer = csv.writer(f, dialect=dialect)
		for row in data:
			writer.writerow(row)

def main():

	datadef = read_ddl(sys.argv[1])
	data, data_rec, dialect = readcsv(sys.argv[2], delete_with_missing=False)
	t_data_rec = fit_to_def(data_rec, datadef)
	t_data = recs_to_rows(t_data_rec)
	writecsv(t_data, sys.argv[3], dialect)

if __name__ == '__main__':
	main()