import pyglet
import cv2
import numpy as np

class Projector(object):

	def __init__(self, screen=0, nchannels=3):

		platform = pyglet.window.get_platform()
		display = platform.get_default_display()
		self.screen = display.get_screens()[screen]

		self.max_width = self.screen.width
		self.max_height = self.screen.height

		self.width = None
		self.height = None

		self.window = pyglet.window.Window(visible=False)
		self.image = None

		self.nchannels = nchannels

	def start_window(self, frame):

		if len(frame.shape) > 2:
			height, width, _ = frame.shape
		else:
			height, width = frame.shape
			frame = cv2.cvtColor(frame, cv2.COLOR_GRAY2BGR)
			self.nchannels = 3

		if height > self.max_height or width > self.max_width:
			if width > height:
				scale_f = float(self.max_width) / float(width)
			else:
				scale_f = float(self.max_height) / float(height)

			height = int(float(height)*scale_f)
			width = int(float(width)*scale_f)

			frame = cv2.resize(frame, (width, height))			

		self.width = width
		self.height = height

		self.image = pyglet.image.ImageData(width, height, 'BGR', frame.tostring(), -1*self.nchannels*width)
		self.image.anchor_x = self.image.width // 2
		self.image.anchor_y = self.image.height // 2

		@self.window.event
		def on_draw():
			self.window.clear()
			self.image.blit(self.screen.width // 2, self.screen.height // 2)
			# self.image.blit(0, 0)

		self.window.set_fullscreen(screen=self.screen)
		self.window.set_visible()

	def project_frame(self, frame):
		
		if len(frame.shape) < 3:
			frame = cv2.cvtColor(frame, cv2.COLOR_GRAY2BGR)

		frame = cv2.resize(frame, (self.width, self.height))
		
		self.image.set_data('BGR', -1*self.nchannels*self.image.width, frame.tostring())
		self.window.clear()
		
		pyglet.clock.tick()		
		self.window.switch_to()
		self.window.dispatch_events()
		self.window.dispatch_event('on_draw')
		self.window.flip()		

