import os
import sys
import json
import tempfile
import pdb
import random

import image_inversion

from multiencoder import EncoderCascade

def run_blur_softmax(model, target_image_file, experiment_dir="../experiments", blackbox=False):

    image_split = os.path.split(os.path.split(target_image_file)[0])
    image_class = image_split[len(image_split)-1]    

    config = {}
    config["model"] = model
    config["target_image"] = target_image_file
    config["starting_image"] = os.path.join(os.path.dirname(target_image_file), "5.pgm")    
    config["img_width"] = 92
    config["img_height"] = 112
    config["img_channels"] = 1
    config["rate"] = 0.1    
    config["patience"] = 1000
    config["random"] = False
    config["blur"] = False
    config["scale"] = 2    
    config["output_dir"] = "%s/%s_bb=%s_%s" % (experiment_dir, os.path.basename(model), str(blackbox).lower(), image_class)
    config["suppress_graphics"] = True
    config["pixelate_width"] = 1

    exp_path = os.path.join("../experiments", config["output_dir"])
    if os.path.exists(exp_path):
        return

    if not os.path.exists("../experiments"):
        os.mkdir("../experiments")

    fp, fname = tempfile.mkstemp(suffix='.json', dir='/tmp')
    os.close(fp)
    with open(fname, "w") as fp:
        json.dump(config, fp, sort_keys=True, indent=4)        

    image_inversion.main(fname)

    os.unlink(fname)

def run_blur_mlp(model, target_image_file, experiment_dir="../experiments", blackbox=False):

    image_split = os.path.split(os.path.split(target_image_file)[0])
    image_class = image_split[len(image_split)-1]    

    config = {}
    config["model"] = model
    config["target_image"] = target_image_file
    config["starting_image"] = os.path.join(os.path.dirname(target_image_file), "5.pgm")    
    config["img_width"] = 92
    config["img_height"] = 112
    config["img_channels"] = 1
    config["rate"] = 0.1    
    config["patience"] = 5000
    config["random"] = False
    config["blur"] = False
    config["scale"] = 2    
    config["output_dir"] = "%s/%s_bb=%s_%s" % (experiment_dir, os.path.basename(model), str(blackbox).lower(), image_class)
    config["suppress_graphics"] = True
    config["pixelate_width"] = 1

    exp_path = os.path.join("../experiments", config["output_dir"])
    if os.path.exists(exp_path):
        return

    if not os.path.exists("../experiments"):
        os.mkdir("../experiments")

    fp, fname = tempfile.mkstemp(suffix='.json', dir='/tmp')
    os.close(fp)
    with open(fname, "w") as fp:
        json.dump(config, fp, sort_keys=True, indent=4)        

    image_inversion.main(fname)

    os.unlink(fname)

def run_blur_dae(model, target_image_file, experiment_dir="../experiments", blackbox=False):

    image_split = os.path.split(os.path.split(target_image_file)[0])
    image_class = image_split[len(image_split)-1]    

    config = {}
    config["model"] = model
    config["target_image"] = target_image_file
    config["starting_image"] = os.path.join(os.path.dirname(target_image_file), "5.pgm")    
    config["img_width"] = 92
    config["img_height"] = 112
    config["img_channels"] = 1
    config["rate"] = 0.1    
    config["patience"] = 1000
    config["random"] = False
    config["blur"] = False
    config["scale"] = 2    
    config["output_dir"] = "%s/%s_bb=%s_%s" % (experiment_dir, os.path.basename(model), str(blackbox).lower(), image_class)
    config["suppress_graphics"] = True
    config["pixelate_width"] = 1
    config["encoder"] = "../models/final/multi_ae_att_full.pkl"

    exp_path = os.path.join("../experiments", config["output_dir"])
    if os.path.exists(exp_path):
        return

    if not os.path.exists("../experiments"):
        os.mkdir("../experiments")

    fp, fname = tempfile.mkstemp(suffix='.json', dir='/tmp')
    os.close(fp)
    with open(fname, "w") as fp:
        json.dump(config, fp, sort_keys=True, indent=4)        

    image_inversion.main(fname)

    os.unlink(fname)


def run_att(model, target_image_file, experiment_dir="../experiments", blackbox=False):

    image_split = os.path.split(os.path.split(target_image_file)[0])
    image_class = image_split[len(image_split)-1]    

    config = {}
    config["model"] = model
    config["target_image"] = target_image_file
    config["img_width"] = 92
    config["img_height"] = 112
    config["img_channels"] = 1
    config["rate"] = 0.1
    config["p_cutoff"] = 0.99
    config["patience"] = 5000
    config["random"] = False
    config["blur"] = False
    config["scale"] = 2    
    config["output_dir"] = "%s/%s_bb=%s_%s" % (experiment_dir, os.path.basename(model), str(blackbox).lower(), image_class)
    config["suppress_graphics"] = True

    if not os.path.exists("../experiments"):
        os.mkdir("../experiments")

    fp, fname = tempfile.mkstemp(suffix='.json', dir='/tmp')
    os.close(fp)
    with open(fname, "w") as fp:
        json.dump(config, fp, sort_keys=True, indent=4)        

    image_inversion.main(fname)

    os.unlink(fname)

def run_att_half_ae_whitebox(model, target_image_file, experiment_dir="../experiments", blackbox=False):

    image_split = os.path.split(os.path.split(target_image_file)[0])
    image_class = image_split[len(image_split)-1]    

    config = {}
    config["model"] = model
    config["target_image"] = target_image_file
    config["img_width"] = 92
    config["img_height"] = 112
    config["img_channels"] = 1
    config["rate"] = 0.25
    config["p_cutoff"] = 0.98
    config["patience"] = 5000
    config["random"] = False
    config["blur"] = False
    config["scale"] = 2    
    config["output_dir"] = "%s/%s_bb=%s_%s" % (experiment_dir, os.path.basename(model), str(blackbox).lower(), image_class)
    config["suppress_graphics"] = True    
    config["encoder"] = "../models/final/multi_ae_att_half.pkl"

    if not os.path.exists("../experiments"):
        os.mkdir("../experiments")

    fp, fname = tempfile.mkstemp(suffix='.json', dir='/tmp')
    os.close(fp)
    with open(fname, "w") as fp:
        json.dump(config, fp, sort_keys=True, indent=4)        

    image_inversion.main(fname)

    os.unlink(fname)    

def run_att_full_ae_whitebox(model, target_image_file, experiment_dir="../experiments", blackbox=False):

    image_split = os.path.split(os.path.split(target_image_file)[0])
    image_class = image_split[len(image_split)-1]    

    config = {}
    config["model"] = model
    config["target_image"] = target_image_file
    config["img_width"] = 92
    config["img_height"] = 112
    config["img_channels"] = 1
    config["rate"] = 0.25
    config["p_cutoff"] = 0.99
    config["patience"] = 5000
    config["random"] = False
    config["blur"] = False
    config["scale"] = 2    
    config["output_dir"] = "%s/%s_bb=%s_%s" % (experiment_dir, os.path.basename(model), str(blackbox).lower(), image_class)
    config["suppress_graphics"] = True    
    config["encoder"] = "../models/final/multi_ae_att_full.pkl"

    if not os.path.exists("../experiments"):
        os.mkdir("../experiments")

    fp, fname = tempfile.mkstemp(suffix='.json', dir='/tmp')
    os.close(fp)
    with open(fname, "w") as fp:
        json.dump(config, fp, sort_keys=True, indent=4)        

    image_inversion.main(fname)

    os.unlink(fname)    


def main():

    random.seed(1)

    path = sys.argv[1]
    ext = sys.argv[2]
    experiment_dir = sys.argv[3]

    models = sys.argv[4:]
    if os.path.basename(models[0]).startswith("dae") and os.path.basename(models[0]).find("half") >= 0:        
        runs = [
            lambda model, target_image_file: run_att_half_ae_whitebox(model, target_image_file, experiment_dir, False),
            lambda model, target_image_file: run_att(model, target_image_file, experiment_dir, True)
        ]
    elif os.path.basename(models[0]).startswith("dae") and os.path.basename(models[0]).find("full") >= 0:
        # runs = [
        #     lambda model, target_image_file: run_att_full_ae_whitebox(model, target_image_file, experiment_dir, False),
        #     # lambda model, target_image_file: run_att(model, target_image_file, experiment_dir, False)
        #     # lambda model, target_image_file: run_att(model, target_image_file, experiment_dir, True)
        # ]
        runs = [
            lambda model, target_image_file: run_blur_dae(model, target_image_file, experiment_dir, False),            
        ]        
    else:    
        # runs = [
        #     lambda model, target_image_file: run_att(model, target_image_file, experiment_dir, False)
        #     # lambda model, target_image_file: run_att(model, target_image_file, experiment_dir, True)
        # ]
        runs = [
            lambda model, target_image_file: run_blur_softmax(model, target_image_file, experiment_dir, False)
            # lambda model, target_image_file: run_blur_mlp(model, target_image_file, experiment_dir, True)
        ]        
    
    for run in runs:
        for model in models:
            for root, dirs, files in os.walk(path):
                ext_files = sorted([curfile for curfile in files if curfile.endswith(ext)], key=lambda x: int(x.split('.')[0]))
                if len(ext_files) == 0:
                    continue                
                target_image_file = os.path.join(root, random.choice(ext_files[7:]))
                print "running %s on %s" % (model, target_image_file)
                run(model, target_image_file)

if __name__ == '__main__':
    main()
