import os
import csv
import sys
import pdb
import argparse

exps = {
	'recon': {
		'blackbox': {
			'softmax': ["../experiments/turkers/softmax-blackbox-50-1.csv"],
			'mlp': ["../experiments/turkers/mlp-blackbox-25-1.csv"],
			'dae': ["../experiments/turkers/dae-blackbox-50-1.csv"]
		},
		'whitebox': {
			'softmax': ["../experiments/turkers/softmax-whitebox-50-1.csv", "../experiments/turkers/softmax-whitebox-50-2.csv", "../experiments/turkers/softmax-whitebox-50-2.csv", "../experiments/turkers/softmax-whitebox-70.csv", "../experiments/turkers/softmax-whitebox-notrue.csv"],
			'mlp': ["../experiments/turkers/mlp-whitebox-25-1.csv", "../experiments/turkers/mlp-whitebox-50-1.csv", "../experiments/turkers/mlp-whitebox-50-2.csv", "../experiments/turkers/mlp-whitebox-50-3.csv"],
			'dae': ["../experiments/turkers/dae-whitebox-50-1.csv", "../experiments/turkers/dae-whitebox-50-2.csv", "../experiments/turkers/dae-whitebox-50-3.csv"]
		},
	},
	'deblur': {
		'plain': {
			'softmax': ["../experiments/turkers/deblur-softmax-50-1.csv", "../experiments/turkers/deblur-softmax-50-2.csv", "../experiments/turkers/deblur-softmax-50-3.csv", "../experiments/turkers/deblur-control-softmax-50-1.csv"],
			'mlp': ["../experiments/turkers/deblur-mlp-50-1.csv", "../experiments/turkers/deblur-mlp-50-2.csv", "../experiments/turkers/deblur-mlp-50-3.csv", "../experiments/turkers/deblur-control-mlp-50-1.csv"],
			'dae': ["../experiments/turkers/deblur-dae-50-1.csv", "../experiments/turkers/deblur-dae-50-2.csv", "../experiments/turkers/deblur-dae-50-3.csv", "../experiments/turkers/deblur-control-dae-50-1.csv"]
		},
		'outside': {
			'softmax': ["../experiments/turkers/deblur-control-softmax-50-1.csv"],
			'mlp': ["../experiments/turkers/deblur-control-mlp-50-1.csv"],
			'dae': ["../experiments/turkers/deblur-control-dae-50-1.csv"]
		}
	}
}

def all_exp_files():

	exp_files = []

	for attack in ['recon', 'deblur']:
		for config in exps[attack].keys():
			for model in ['softmax', 'mlp', 'dae']:
				exp_files.extend(exps[attack][config][model])

	return exp_files

def get_csv(filen):

	data = []
	
	with open(filen, 'rU') as f:
		reader = csv.reader(f, delimiter = ",", quotechar = "\"")
		attrline = reader.next()
		attrline = [el.strip() for el in attrline]
		features = attrline[:]
		attr_indices = {attrline[i]: i for i in range(len(attrline))}
		attributes = {i: attrline[i] for i in range(len(attrline))}
		
		for row in reader:
			drow = {attributes[i]: row[i] for i in range(len(row))}
			# fix up coding error from first batch
			if int(drow['Answer.guess']) == -1:
				drow['Answer.guess'] = "0"
			# fix up coding error from deblur control group
			if filen.find('deblur-control') != -1:
				drow['Input.true_idx'] = "0"
			# fix up coding error from deblur main group
			if filen.find('deblur') != -1 and filen.find('control') == -1 and int(drow['Input.true_idx']) == 0:
				continue
			data.append(drow)

	return data

def get_workers(tabs):

	workers = set()
	for data in tabs:
		for row in data:
			workers.add(row['WorkerId'])

	return list(workers)

def worker_correct(data, workerid):

	n = 0
	for row in data:
		if row['WorkerId'] == workerid and row['Input.true_idx'] == row['Answer.guess']:
			n += 1

	return n

def worker_answered(data, workerid):
	n = 0
	for row in data:
		if row['WorkerId'] == workerid:
			n += 1

	return n

def count_hits():

	all_files = all_exp_files()
	tabs = [get_csv(filen) for filen in all_files]

	return sum([len(data) for data in tabs])

def worker_stats(workers=None):

	all_files = all_exp_files()

	tabs = [get_csv(filen) for filen in all_files]	

	if workers is None:
		workers = get_workers(tabs)
	worker_crct = {worker: 0 for worker in workers}
	worker_ans = {worker: 0 for worker in workers}
	worker_pct = {worker: 0.0 for worker in workers}
	for worker in workers:
		for data in tabs:
			worker_crct[worker] += worker_correct(data, worker)
			worker_ans[worker] += worker_answered(data, worker)			
	for worker in workers:
		worker_pct[worker] = float(worker_crct[worker])/float(worker_ans[worker])

	return workers, worker_crct, worker_ans, worker_pct

def select_workers(n_completed=0, min_correct=0.0):

	workers, correct, answered, pct = worker_stats()

	good_workers = []
	for worker in workers:
		if answered[worker] >= n_completed and pct[worker] >= min_correct:
			good_workers.append(worker)

	return good_workers

def discretize_pct(pct):	
	for i in range(1,11):
		low = (i-1)
		if pct > i*0.1-0.09 and pct <= i*0.1:
			return "[%d-%d]" % (100*i*(0.1-0.09), 100*i*0.1)
	return "[0]"

def worker_stats_csv(out_filen, workers=None):	

	example_csv = all_exp_files()[0]
	with open(example_csv, 'rU') as f:
		reader = csv.reader(f, delimiter = ",", quotechar = "\"")
		attrline = reader.next()				

	with open(out_filen, 'wb') as f:
		writer = csv.writer(f)
		writer.writerow(["worker_id", "correct", "answered", "percent", "range"])

		workers, worker_crct, worker_ans, worker_pct = worker_stats(workers)
		for worker in workers:
			writer.writerow([worker, worker_crct[worker], worker_ans[worker], worker_pct[worker], discretize_pct(worker_pct[worker])])

def tabs_with_workers(tabs, workers):
	
	new_tabs = []
	for data in tabs:
		new_data = []
		for row in data:
			if row['WorkerId'] in workers:
				new_data.append(row)
		if len(new_data) > 0:
			new_tabs.append(new_data)

	return new_tabs

def cond_correct(row):
	return True

def cond_identified(row):
	return int(row['Input.true_idx']) != 0

def cond_excluded(row):
	return int(row['Input.true_idx']) == 0

def cmp_correct(row):
	return 1 if cond_correct(row) and row['Input.true_idx'] == row['Answer.guess'] else 0

def cmp_identified(row):	
	return 1 if cond_identified(row) and row['Input.true_idx'] == row['Answer.guess'] else 0

def cmp_excluded(row):
	return 1 if cond_excluded(row) and row['Input.true_idx'] == row['Answer.guess'] else 0

def main_stats(tabs, control=False):

	correct = 0.0
	identified = 0.0
	excluded = 0.0
	correct_total = 0.0
	identified_total = 0.0
	excluded_total = 0.0

	for data in tabs:
		for row in data:			
			if row['Input.control'] != str(control).lower():
				continue
			correct += cmp_correct(row)
			identified += cmp_identified(row)
			excluded += cmp_excluded(row)
			correct_total += 1 if cond_correct(row) else 0		
			identified_total += 1 if cond_identified(row) else 0
			excluded_total += 1 if cond_excluded(row) else 0

	correct /= correct_total if correct_total > 0 else 1.0
	identified /= identified_total if identified_total > 0 else 1.0
	excluded /= excluded_total if excluded_total > 0 else 1.0

	return correct, identified, excluded

def hit_str(row):

	return "%s%s%s%s%s%s" % (row["Input.inverted_face"], row["Input.test1"], row["Input.test2"], row["Input.test3"], row["Input.test4"], row["Input.test5"])

def row_answers(row):

	return "true: %s, guess: %s" % (row['Input.true_idx'], row['Answer.guess'])

def majority_stats(tabs, min_majority = 0.51, control=False):

	hits = {}
	for data in tabs:
		for row in data:
			if row['Input.control'] != str(control).lower():
				continue
			hitstr = hit_str(row)
			if not hitstr in hits.keys():
				hits[hitstr] = {'correct': 0, 'identified': 0, 'excluded': 0, 'correct_cnt': 0, 'identified_cnt': 0, 'excluded_cnt': 0}
			hits[hitstr]['correct'] += cmp_correct(row)
			hits[hitstr]['identified'] += cmp_identified(row)
			hits[hitstr]['excluded'] += cmp_excluded(row)
			hits[hitstr]['correct_cnt'] += 1 if cond_correct(row) else 0
			hits[hitstr]['identified_cnt'] += 1 if cond_identified(row) else 0
			hits[hitstr]['excluded_cnt'] += 1 if cond_excluded(row) else 0

	for hit in hits.keys():
		if hits[hit]['correct_cnt'] > 1:
			hits[hit]['correct'] = 1 if float(hits[hit]['correct'])/float(hits[hit]['correct_cnt']) >= min_majority else 0
		else:
			hits[hit]['correct'] = 0
		if hits[hit]['identified_cnt'] > 1:
			hits[hit]['identified'] = 1 if float(hits[hit]['identified'])/float(hits[hit]['identified_cnt']) >= min_majority else 0
		else:
			hits[hit]['identified'] = 0
		if hits[hit]['excluded_cnt'] > 1:
			hits[hit]['excluded'] = 1 if float(hits[hit]['excluded'])/float(hits[hit]['excluded_cnt']) >= min_majority else 0
		else:
			hits[hit]['excluded'] = 0

	correct_cnt = sum([1 if hits[hit]['correct_cnt'] > 1 else 0 for hit in hits.keys()])
	identified_cnt = sum([1 if hits[hit]['identified_cnt'] > 1 else 0 for hit in hits.keys()])
	excluded_cnt = sum([1 if hits[hit]['excluded_cnt'] > 1 else 0 for hit in hits.keys()])
	
	if correct_cnt > 0:
		correct = float(sum([hits[hit]['correct'] for hit in hits.keys()]))/correct_cnt
	else:
		correct = 0
	if identified_cnt > 0:
		identified = float(sum([hits[hit]['identified'] for hit in hits.keys()]))/identified_cnt
	else:
		identified = 0
	if excluded_cnt > 0:
		excluded = float(sum([hits[hit]['excluded'] for hit in hits.keys()]))/excluded_cnt
	else:
		excluded = 0

	return correct, identified, excluded

def experiment_stats(exp_files, workers=None, control=False):

	tabs = []
	for filen in exp_files:
		tabs.append(get_csv(filen))

	if workers is None:
		workers = get_workers(tabs)

	tabs = tabs_with_workers(tabs, workers)

	correct, identified, excluded = main_stats(tabs, control=control)
	correct_m, identified_m, excluded_m = majority_stats(tabs, control=control)

	print "\tcorrect: %.3f" % correct, "(majority: %.3f)" % correct_m
	print "\tidentified: %.3f" % identified, "(majority: %.3f)" % identified_m
	print "\texcluded: %.3f" % excluded, "(majority: %.3f)" % excluded_m

	return correct, identified, excluded

def process_args():
    parser = argparse.ArgumentParser(description='Turk experiment tabulation.')
    parser.add_argument('--n_completed', type=int, default=0, help='minimum number of tasks completed by workers')
    parser.add_argument('--min_pct', type=float, default=0.0, help='minimum worker accuracy')
    parser.add_argument('--worker_stats', type=str, default=None, help='output worker stats to file')
    parser.add_argument('--do_recon', action='store_true', help='tabulate reconstruction attacks')
    parser.add_argument('--do_blur', action='store_true', help='tabulate deblurring attacks')
    parser.add_argument('--count_hits', action='store_true', help='give total HIT count')
    parser.add_argument('--control', action='store_true', help='run on control group')    

    return parser.parse_args()

def main():

	args = process_args()

	n_completed = args.n_completed
	min_pct = args.min_pct
	workers = select_workers(n_completed, min_pct)

	attacks = []
	if args.do_recon:
		attacks.append('recon')
	if args.do_blur:
		attacks.append('deblur')

	print "\n...using %d workers\n" % len(workers)

	results = []

	for attack in attacks:
		for config in exps[attack].keys():
			for model in ['softmax', 'mlp', 'dae']:
				print "%s-%s-%s" % (attack, config, model)			
				results.append(experiment_stats(exps[attack][config][model], workers, args.control))
				print

	if args.control:		
		print "mean control accuracy: %.3f" % (float(sum([float(r[0]) for r in results]))/float(len(results)))

	if args.count_hits:
		print "total HITs:", count_hits()

	if args.worker_stats is not None:
		worker_stats_csv(args.worker_stats, workers)


if __name__ == "__main__":

	main()